@@ -1,14 +1,15 @@
 import os.path as op
+import warnings
 
 from functools import wraps
 
 from flask import Blueprint, current_app, render_template, abort, g, url_for
-from flask.ext.admin import babel
-from flask.ext.admin._compat import with_metaclass
-from flask.ext.admin import helpers as h
+from flask_admin import babel
+from flask_admin._compat import with_metaclass, as_unicode
+from flask_admin import helpers as h
 
 # For compatibility reasons import MenuLink
-from flask.ext.admin.menu import MenuCategory, MenuView, MenuLink
+from flask_admin.menu import MenuCategory, MenuView, MenuLink, SubMenuCategory  # noqa: F401
 
 
 def expose(url='/', methods=('GET',)):
@@ -111,11 +112,15 @@ class BaseView(with_metaclass(AdminViewMeta, BaseViewClass)):
 
         Derive from this class to implement your administrative interface piece. For example::
 
-            from flask.ext.admin import BaseView, expose
+            from flask_admin import BaseView, expose
             class MyView(BaseView):
                 @expose('/')
                 def index(self):
                     return 'Hello World!'
+
+        Icons can be added to the menu by using `menu_icon_type` and `menu_icon_value`. For example::
+
+            admin.add_view(MyView(name='My View', menu_icon_type='glyph', menu_icon_value='glyphicon-home'))
     """
     @property
     def _template_args(self):
@@ -175,15 +180,16 @@ class BaseView(with_metaclass(AdminViewMeta, BaseViewClass)):
             :param menu_icon_type:
                 Optional icon. Possible icon types:
 
-                 - `flask.ext.admin.consts.ICON_TYPE_GLYPH` - Bootstrap glyph icon
-                 - `flask.ext.admin.consts.ICON_TYPE_IMAGE` - Image relative to Flask static directory
-                 - `flask.ext.admin.consts.ICON_TYPE_IMAGE_URL` - Image with full URL
+                 - `flask_admin.consts.ICON_TYPE_GLYPH` - Bootstrap glyph icon
+                 - `flask_admin.consts.ICON_TYPE_FONT_AWESOME` - Font Awesome icon
+                 - `flask_admin.consts.ICON_TYPE_IMAGE` - Image relative to Flask static directory
+                 - `flask_admin.consts.ICON_TYPE_IMAGE_URL` - Image with full URL
             :param menu_icon_value:
                 Icon glyph name or URL, depending on `menu_icon_type` setting
         """
         self.name = name
         self.category = category
-        self.endpoint = endpoint
+        self.endpoint = self._get_endpoint(endpoint)
         self.url = url
         self.static_folder = static_folder
         self.static_url_path = static_url_path
@@ -201,6 +207,34 @@ class BaseView(with_metaclass(AdminViewMeta, BaseViewClass)):
         if self._default_view is None:
             raise Exception(u'Attempted to instantiate admin view %s without default view' % self.__class__.__name__)
 
+    def _get_endpoint(self, endpoint):
+        """
+            Generate Flask endpoint name. By default converts class name to lower case if endpoint is
+            not explicitly provided.
+        """
+        if endpoint:
+            return endpoint
+
+        return self.__class__.__name__.lower()
+
+    def _get_view_url(self, admin, url):
+        """
+            Generate URL for the view. Override to change default behavior.
+        """
+        if url is None:
+            if admin.url != '/':
+                url = '%s/%s' % (admin.url, self.endpoint)
+            else:
+                if self == admin.index_view:
+                    url = '/'
+                else:
+                    url = '/%s' % self.endpoint
+        else:
+            if not url.startswith('/'):
+                url = '%s/%s' % (admin.url, url)
+
+        return url
+
     def create_blueprint(self, admin):
         """
             Create Flask blueprint.
@@ -208,32 +242,22 @@ class BaseView(with_metaclass(AdminViewMeta, BaseViewClass)):
         # Store admin instance
         self.admin = admin
 
-        # If endpoint name is not provided, get it from the class name
-        if self.endpoint is None:
-            self.endpoint = self.__class__.__name__.lower()
-
         # If the static_url_path is not provided, use the admin's
         if not self.static_url_path:
             self.static_url_path = admin.static_url_path
 
-        # If url is not provided, generate it from endpoint name
-        if self.url is None:
-            if self.admin.url != '/':
-                self.url = '%s/%s' % (self.admin.url, self.endpoint)
-            else:
-                if self == admin.index_view:
-                    self.url = '/'
-                else:
-                    self.url = '/%s' % self.endpoint
-        else:
-            if not self.url.startswith('/'):
-                self.url = '%s/%s' % (self.admin.url, self.url)
+        # Generate URL
+        self.url = self._get_view_url(admin, self.url)
 
         # If we're working from the root of the site, set prefix to None
         if self.url == '/':
             self.url = None
+            # prevent admin static files from conflicting with flask static files
+            if not self.static_url_path:
+                self.static_folder = 'static'
+                self.static_url_path = '/static/admin'
 
-        # If name is not povided, use capitalized endpoint name
+        # If name is not provided, use capitalized endpoint name
         if self.name is None:
             self.name = self._prettify_class_name(self.__class__.__name__)
 
@@ -383,10 +407,22 @@ class AdminIndexView(BaseView):
                 @expose('/')
                 def index(self):
                     arg1 = 'Hello'
-                    return render_template('adminhome.html', arg1=arg1)
+                    return self.render('admin/myhome.html', arg1=arg1)
 
             admin = Admin(index_view=MyHomeView())
 
+
+        Also, you can change the root url from /admin to / with the following::
+
+            admin = Admin(
+                app,
+                index_view=AdminIndexView(
+                    name='Home',
+                    template='admin/myhome.html',
+                    url='/'
+                )
+            )
+
         Default values for the index page are:
 
         * If a name is not provided, 'Home' will be used.
@@ -397,12 +433,18 @@ class AdminIndexView(BaseView):
     """
     def __init__(self, name=None, category=None,
                  endpoint=None, url=None,
-                 template='admin/index.html'):
+                 template='admin/index.html',
+                 menu_class_name=None,
+                 menu_icon_type=None,
+                 menu_icon_value=None):
         super(AdminIndexView, self).__init__(name or babel.lazy_gettext('Home'),
                                              category,
                                              endpoint or 'admin',
-                                             url or '/admin',
-                                             'static')
+                                             '/admin' if url is None else url,
+                                             'static',
+                                             menu_class_name=menu_class_name,
+                                             menu_icon_type=menu_icon_type,
+                                             menu_icon_value=menu_icon_value)
         self._template = template
 
     @expose()
@@ -421,7 +463,8 @@ class Admin(object):
                  endpoint=None,
                  static_url_path=None,
                  base_template=None,
-                 template_mode=None):
+                 template_mode=None,
+                 category_icon_classes=None):
         """
             Constructor.
 
@@ -448,7 +491,10 @@ class Admin(object):
                 Override base HTML template for all static views. Defaults to `admin/base.html`.
             :param template_mode:
                 Base template path. Defaults to `bootstrap2`. If you want to use
-                Bootstrap 3 integration, change it to `bootstrap3`.
+                Bootstrap 3 or 4 integration, change it to `bootstrap3` or `bootstrap4`.
+            :param category_icon_classes:
+                A dict of category names as keys and html classes as values to be added to menu category icons.
+                Example: {'Favorites': 'glyphicon glyphicon-star'}
         """
         self.app = app
 
@@ -470,9 +516,10 @@ class Admin(object):
         self.subdomain = subdomain
         self.base_template = base_template or 'admin/base.html'
         self.template_mode = template_mode or 'bootstrap2'
+        self.category_icon_classes = category_icon_classes or dict()
 
-        # Add predefined index view
-        self.add_view(self.index_view)
+        # Add index view
+        self._set_admin_index_view(index_view=index_view, endpoint=endpoint, url=url)
 
         # Register with application
         if app is not None:
@@ -494,6 +541,68 @@ class Admin(object):
 
         self._add_view_to_menu(view)
 
+    def _set_admin_index_view(self, index_view=None,
+                              endpoint=None, url=None):
+        """
+            Add the admin index view.
+
+          :param index_view:
+               Home page view to use. Defaults to `AdminIndexView`.
+           :param url:
+               Base URL
+          :param endpoint:
+               Base endpoint name for index view. If you use multiple instances of the `Admin` class with
+               a single Flask application, you have to set a unique endpoint name for each instance.
+        """
+        self.index_view = index_view or AdminIndexView(endpoint=endpoint, url=url)
+        self.endpoint = endpoint or self.index_view.endpoint
+        self.url = url or self.index_view.url
+
+        # Add predefined index view
+        # assume index view is always the first element of views.
+        if len(self._views) > 0:
+            self._views[0] = self.index_view
+            self._menu[0] = MenuView(self.index_view.name, self.index_view)
+        else:
+            self.add_view(self.index_view)
+
+    def add_views(self, *args):
+        """
+            Add one or more views to the collection.
+
+            Examples::
+
+                admin.add_views(view1)
+                admin.add_views(view1, view2, view3, view4)
+                admin.add_views(*my_list)
+
+            :param args:
+                Argument list including the views to add.
+        """
+        for view in args:
+            self.add_view(view)
+
+    def add_sub_category(self, name, parent_name):
+
+        """
+            Add a category of a given name underneath
+            the category with parent_name.
+
+            :param name:
+                The name of the new menu category.
+            :param parent_name:
+                The name of a parent_name category
+        """
+
+        name_text = as_unicode(name)
+        parent_name_text = as_unicode(parent_name)
+        category = self.get_category_menu_item(name_text)
+        parent = self.get_category_menu_item(parent_name_text)
+        if category is None and parent is not None:
+            category = SubMenuCategory(name)
+            self._menu_categories[name_text] = category
+            parent.add_child(category)
+
     def add_link(self, link):
         """
             Add link to menu links collection.
@@ -502,23 +611,45 @@ class Admin(object):
                 Link to add.
         """
         if link.category:
-            self._add_menu_item(link, link.category)
+            self.add_menu_item(link, link.category)
         else:
             self._menu_links.append(link)
 
-    def _add_menu_item(self, menu_item, target_category):
+    def add_links(self, *args):
         """
-            Add a view to the menu tree
+            Add one or more links to the menu links collection.
 
-            :param view:
-                View to add
+            Examples::
+
+                admin.add_links(link1)
+                admin.add_links(link1, link2, link3, link4)
+                admin.add_links(*my_list)
+
+            :param args:
+                Argument list including the links to add.
+        """
+        for link in args:
+            self.add_link(link)
+
+    def add_menu_item(self, menu_item, target_category=None):
+        """
+            Add menu item to menu tree hierarchy.
+
+            :param menu_item:
+                MenuItem class instance
+            :param target_category:
+                Target category name
         """
         if target_category:
-            category = self._menu_categories.get(target_category)
+            cat_text = as_unicode(target_category)
 
+            category = self._menu_categories.get(cat_text)
+
+            # create a new menu category if one does not exist already
             if category is None:
                 category = MenuCategory(target_category)
-                self._menu_categories[target_category] = category
+                category.class_name = self.category_icon_classes.get(cat_text)
+                self._menu_categories[cat_text] = category
 
                 self._menu.append(category)
 
@@ -526,13 +657,24 @@ class Admin(object):
         else:
             self._menu.append(menu_item)
 
+    def _add_menu_item(self, menu_item, target_category):
+        warnings.warn('Admin._add_menu_item is obsolete - use Admin.add_menu_item instead.')
+        return self.add_menu_item(menu_item, target_category)
+
     def _add_view_to_menu(self, view):
-        self._add_menu_item(MenuView(view.name, view), view.category)
+        """
+            Add a view to the menu tree
+
+            :param view:
+                View to add
+        """
+        self.add_menu_item(MenuView(view.name, view), view.category)
 
     def get_category_menu_item(self, name):
         return self._menu_categories.get(name)
 
-    def init_app(self, app):
+    def init_app(self, app, index_view=None,
+                 endpoint=None, url=None):
         """
             Register all views with the Flask application.
 
@@ -543,6 +685,14 @@ class Admin(object):
 
         self._init_extension()
 
+        # Register Index view
+        if index_view is not None:
+            self._set_admin_index_view(
+                index_view=index_view,
+                endpoint=endpoint,
+                url=url
+            )
+
         # Register views
         for view in self._views:
             app.register_blueprint(view.create_blueprint(self))
