@@ -1,21 +1,22 @@
 import os
 import os.path as op
 
-from werkzeug import secure_filename
+from werkzeug.utils import secure_filename
 from werkzeug.datastructures import FileStorage
 
 from wtforms import ValidationError, fields
-from wtforms.widgets import HTMLString, html_params
+from wtforms.widgets import html_params
 
 try:
     from wtforms.fields.core import _unset_value as unset_value
 except ImportError:
     from wtforms.utils import unset_value
 
-from flask.ext.admin.babel import gettext
-from flask.ext.admin.helpers import get_url
+from flask_admin.babel import gettext
+from flask_admin.helpers import get_url
 
-from flask.ext.admin._compat import string_types, urljoin
+from flask_admin._backwards import Markup
+from flask_admin._compat import string_types, urljoin
 
 
 try:
@@ -38,6 +39,7 @@ class FileUploadInput(object):
         look and feel.
     """
     empty_template = ('<input %(file)s>')
+    input_type = 'file'
 
     data_template = ('<div>'
                      ' <input %(text)s>'
@@ -51,11 +53,21 @@ class FileUploadInput(object):
 
         template = self.data_template if field.data else self.empty_template
 
-        return HTMLString(template % {
+        if field.errors:
+            template = self.empty_template
+
+        if field.data and isinstance(field.data, FileStorage):
+            value = field.data.filename
+        else:
+            value = field.data or ''
+
+        return Markup(template % {
             'text': html_params(type='text',
                                 readonly='readonly',
-                                value=field.data),
+                                value=value,
+                                name=field.name),
             'file': html_params(type='file',
+                                value=value,
                                 **kwargs),
             'marker': '_%s-delete' % field.name
         })
@@ -69,10 +81,12 @@ class ImageUploadInput(object):
         look and feel.
     """
     empty_template = ('<input %(file)s>')
+    input_type = 'file'
 
     data_template = ('<div class="image-thumbnail">'
                      ' <img %(image)s>'
                      ' <input type="checkbox" name="%(marker)s">Delete</input>'
+                     ' <input %(text)s>'
                      '</div>'
                      '<input %(file)s>')
 
@@ -81,6 +95,9 @@ class ImageUploadInput(object):
         kwargs.setdefault('name', field.name)
 
         args = {
+            'text': html_params(type='hidden',
+                                value=field.data,
+                                name=field.name),
             'file': html_params(type='file',
                                 **kwargs),
             'marker': '_%s-delete' % field.name
@@ -94,7 +111,7 @@ class ImageUploadInput(object):
         else:
             template = self.empty_template
 
-        return HTMLString(template % args)
+        return Markup(template % args)
 
     def get_url(self, field):
         if field.thumbnail_size:
@@ -121,7 +138,7 @@ class FileUploadField(fields.StringField):
     def __init__(self, label=None, validators=None,
                  base_path=None, relative_path=None,
                  namegen=None, allowed_extensions=None,
-                 permission=0o666,
+                 permission=0o666, allow_overwrite=True,
                  **kwargs):
         """
             Constructor.
@@ -153,6 +170,11 @@ class FileUploadField(fields.StringField):
 
             :param allowed_extensions:
                 List of allowed extensions. If not provided, will allow any file.
+            :param allow_overwrite:
+                Whether to overwrite existing files in upload directory. Defaults to `True`.
+
+            .. versionadded:: 1.1.1
+                The `allow_overwrite` parameter was added.
         """
         self.base_path = base_path
         self.relative_path = relative_path
@@ -160,6 +182,7 @@ class FileUploadField(fields.StringField):
         self.namegen = namegen or namegen_filename
         self.allowed_extensions = allowed_extensions
         self.permission = permission
+        self._allow_overwrite = allow_overwrite
 
         self._should_delete = False
 
@@ -179,13 +202,20 @@ class FileUploadField(fields.StringField):
                 filename.rsplit('.', 1)[1].lower() in
                 map(lambda x: x.lower(), self.allowed_extensions))
 
+    def _is_uploaded_file(self, data):
+        return (data and isinstance(data, FileStorage) and data.filename)
+
     def pre_validate(self, form):
-        if (self.data
-                and self.data.filename
-                and isinstance(self.data, FileStorage)
-                and not self.is_file_allowed(self.data.filename)):
+        if self._is_uploaded_file(self.data) and not self.is_file_allowed(self.data.filename):
             raise ValidationError(gettext('Invalid file extension'))
 
+        # Handle overwriting existing content
+        if not self._is_uploaded_file(self.data):
+            return
+
+        if not self._allow_overwrite and os.path.exists(self._get_path(self.data.filename)):
+            raise ValidationError(gettext('File "%s" already exists.' % self.data.filename))
+
     def process(self, formdata, data=unset_value):
         if formdata:
             marker = '_%s-delete' % self.name
@@ -194,6 +224,15 @@ class FileUploadField(fields.StringField):
 
         return super(FileUploadField, self).process(formdata, data)
 
+    def process_formdata(self, valuelist):
+        if self._should_delete:
+            self.data = None
+        elif valuelist:
+            for data in valuelist:
+                if self._is_uploaded_file(data):
+                    self.data = data
+                    break
+
     def populate_obj(self, obj, name):
         field = getattr(obj, name, None)
         if field:
@@ -203,7 +242,7 @@ class FileUploadField(fields.StringField):
                 setattr(obj, name, None)
                 return
 
-        if self.data and self.data.filename and isinstance(self.data, FileStorage):
+        if self._is_uploaded_file(self.data):
             if field:
                 self._delete_file(field)
 
@@ -241,6 +280,9 @@ class FileUploadField(fields.StringField):
         if not op.exists(op.dirname(path)):
             os.makedirs(os.path.dirname(path), self.permission | 0o111)
 
+        if (self._allow_overwrite is False) and os.path.exists(path):
+            raise ValueError(gettext('File "%s" already exists.' % path))
+
         data.save(path)
 
         return filename
@@ -299,10 +341,13 @@ class ImageUploadField(FileUploadField):
                         upload = FileUploadField('File', namegen=prefix_name)
 
             :param allowed_extensions:
-                List of allowed extensions. If not provided, will allow any file.
+                List of allowed extensions. If not provided, then gif, jpg, jpeg, png and tiff will be allowed.
             :param max_size:
                 Tuple of (width, height, force) or None. If provided, Flask-Admin will
                 resize image to the desired size.
+
+                Width and height is in pixels. If `force` is set to `True`, will try to fit image into dimensions and
+                keep aspect ratio, otherwise will just resize to target size.
             :param thumbgen:
                 Thumbnail filename generation function. All thumbnails will be saved as JPEG files,
                 so there's no need to keep original file extension.
@@ -316,7 +361,7 @@ class ImageUploadField(FileUploadField):
                         return secure_filename('%s-thumb.jpg' % name)
 
                     class MyForm(BaseForm):
-                        upload = ImageUploadField('File', thumbgen=prefix_name)
+                        upload = ImageUploadField('File', thumbgen=thumb_name)
 
             :param thumbnail_size:
                 Tuple or (width, height, force) values. If not provided, thumbnail won't be created.
@@ -357,9 +402,7 @@ class ImageUploadField(FileUploadField):
     def pre_validate(self, form):
         super(ImageUploadField, self).pre_validate(form)
 
-        if (self.data and
-                isinstance(self.data, FileStorage) and
-                self.data.filename):
+        if self._is_uploaded_file(self.data):
             try:
                 self.image = Image.open(self.data)
             except Exception as e:
@@ -382,7 +425,7 @@ class ImageUploadField(FileUploadField):
         path = self._get_path(filename)
 
         if not op.exists(op.dirname(path)):
-            os.makedirs(os.path.dirname(path), self.permission)
+            os.makedirs(os.path.dirname(path), self.permission | 0o111)
 
         # Figure out format
         filename, format = self._get_save_format(filename, self.image)
@@ -396,7 +439,7 @@ class ImageUploadField(FileUploadField):
             self._save_image(image, self._get_path(filename), format)
         else:
             data.seek(0)
-            data.save( self._get_path(filename) )
+            data.save(self._get_path(filename))
 
         self._save_thumbnail(data, filename, format)
 
@@ -424,7 +467,10 @@ class ImageUploadField(FileUploadField):
         return image
 
     def _save_image(self, image, path, format='JPEG'):
-        if image.mode not in ('RGB', 'RGBA'):
+        # New Pillow versions require RGB format for JPEGs
+        if format == 'JPEG' and image.mode != 'RGB':
+            image = image.convert('RGB')
+        elif image.mode not in ('RGB', 'RGBA'):
             image = image.convert('RGBA')
 
         with open(path, 'wb') as fp:
